<?php

/**
* Fast, light and safe Cache Class
*
* Cache_Lite is a fast, light and safe cache system. It's optimized
* for file containers. It is fast and safe (because it uses file
* locking and/or anti-corruption tests).
*
* There are some examples in the 'docs/examples' file
* Technical choices are described in the 'docs/technical' file
*
* Memory Caching is from an original idea of
* Mike BENOIT <ipso@snappymail.ca>
*
* Nota : A chinese documentation (thanks to RainX <china_1982@163.com>) is
* available at :
* http://rainx.phpmore.com/manual/cache_lite.html
*
* @package Cache_Lite
* @category Caching
* @author Fabien MARTY <fab@php.net>
* @author Markus Tacker <tacker@php.net>
*/

define('CACHE_LITE_ERROR_RETURN', 1);
define('CACHE_LITE_ERROR_DIE', 8);

class Cache_Lite
{

    // --- Private properties ---

    /**
    * Directory where to put the cache files
    * (make sure to add a trailing slash)
    *
    * @var string $_cacheDir
    */
    var $_cacheDir = '/tmp/';

    /**
    * Enable / disable caching
    *
    * (can be very usefull for the debug of cached scripts)
    *
    * @var boolean $_caching
    */
    var $_caching = true;

    /**
    * Cache lifetime (in seconds)
    *
    * If null, the cache is valid forever.
    *
    * @var int $_lifeTime
    */
    var $_lifeTime = 3600;

    /**
    * Enable / disable fileLocking
    *
    * (can avoid cache corruption under bad circumstances)
    *
    * @var boolean $_fileLocking
    */
    var $_fileLocking = true;

    /**
    * Timestamp of the last valid cache
    *
    * @var int $_refreshTime
    */
    var $_refreshTime;

    /**
    * File name (with path)
    *
    * @var string $_file
    */
    var $_file;

    /**
    * File name (without path)
    *
    * @var string $_fileName
    */
    var $_fileName;

    /**
    * Enable / disable write control (the cache is read just after writing to detect corrupt entries)
    *
    * Enable write control will lightly slow the cache writing but not the cache reading
    * Write control can detect some corrupt cache files but maybe it's not a perfect control
    *
    * @var boolean $_writeControl
    */
    var $_writeControl = true;

    /**
    * Enable / disable read control
    *
    * If enabled, a control key is embeded in cache file and this key is compared with the one
    * calculated after the reading.
    *
    * @var boolean $_writeControl
    */
    var $_readControl = true;

    /**
    * Type of read control (only if read control is enabled)
    *
    * Available values are :
    * 'md5' for a md5 hash control (best but slowest)
    * 'crc32' for a crc32 hash control (lightly less safe but faster, better choice)
    * 'strlen' for a length only test (fastest)
    *
    * @var boolean $_readControlType
    */
    var $_readControlType = 'crc32';

    /**
    * Pear error mode (when raiseError is called)
    *
    * (see PEAR doc)
    *
    * @see setToDebug()
    * @var int $_pearErrorMode
    */
    var $_pearErrorMode = CACHE_LITE_ERROR_RETURN;

    /**
    * Current cache id
    *
    * @var string $_id
    */
    var $_id;

    /**
    * Current cache group
    *
    * @var string $_group
    */
    var $_group;

    /**
    * Enable / Disable "Memory Caching"
    *
    * NB : There is no lifetime for memory caching !
    *
    * @var boolean $_memoryCaching
    */
    var $_memoryCaching = false;

    /**
    * Enable / Disable "Only Memory Caching"
    * (be carefull, memory caching is "beta quality")
    *
    * @var boolean $_onlyMemoryCaching
    */
    var $_onlyMemoryCaching = false;

    /**
    * Memory caching array
    *
    * @var array $_memoryCachingArray
    */
    var $_memoryCachingArray = array();

    /**
    * Memory caching counter
    *
    * @var int $memoryCachingCounter
    */
    var $_memoryCachingCounter = 0;

    /**
    * Memory caching limit
    *
    * @var int $memoryCachingLimit
    */
    var $_memoryCachingLimit = 1000;

    /**
    * File Name protection
    *
    * if set to true, you can use any cache id or group name
    * if set to false, it can be faster but cache ids and group names
    * will be used directly in cache file names so be carefull with
    * special characters...
    *
    * @var boolean $fileNameProtection
    */
    var $_fileNameProtection = true;

    /**
    * Enable / disable automatic serialization
    *
    * it can be used to save directly datas which aren't strings
    * (but it's slower)
    *
    * @var boolean $_serialize
    */
    var $_automaticSerialization = false;

    /**
    * Disable / Tune the automatic cleaning process
    *
    * The automatic cleaning process destroy too old (for the given life time)
    * cache files when a new cache file is written.
    * 0               => no automatic cache cleaning
    * 1               => systematic cache cleaning
    * x (integer) > 1 => automatic cleaning randomly 1 times on x cache write
    *
    * @var int $_automaticCleaning
    */
    var $_automaticCleaningFactor = 0;

    /**
    * Nested directory level
    *
    * Set the hashed directory structure level. 0 means "no hashed directory
    * structure", 1 means "one level of directory", 2 means "two levels"...
    * This option can speed up Cache_Lite only when you have many thousands of
    * cache file. Only specific benchs can help you to choose the perfect value
    * for you. Maybe, 1 or 2 is a good start.
    *
    * @var int $_hashedDirectoryLevel
    */
    var $_hashedDirectoryLevel = 0;

    /**
    * Umask for hashed directory structure
    *
    * @var int $_hashedDirectoryUmask
    */
    var $_hashedDirectoryUmask = 0700;

    /**
     * API break for error handling in CACHE_LITE_ERROR_RETURN mode
     *
     * In CACHE_LITE_ERROR_RETURN mode, error handling was not good because
     * for example save() method always returned a boolean (a PEAR_Error object
     * would be better in CACHE_LITE_ERROR_RETURN mode). To correct this without
     * breaking the API, this option (false by default) can change this handling.
     *
     * @var boolean
     */
    var $_errorHandlingAPIBreak = false;

    // --- Public methods ---

    /**
    * Constructor
    *
    * $options is an assoc. Available options are :
    * $options = array(
    *     'cacheDir' => directory where to put the cache files (string),
    *     'caching' => enable / disable caching (boolean),
    *     'lifeTime' => cache lifetime in seconds (int),
    *     'fileLocking' => enable / disable fileLocking (boolean),
    *     'writeControl' => enable / disable write control (boolean),
    *     'readControl' => enable / disable read control (boolean),
    *     'readControlType' => type of read control 'crc32', 'md5', 'strlen' (string),
    *     'pearErrorMode' => pear error mode (when raiseError is called) (cf PEAR doc) (int),
    *     'memoryCaching' => enable / disable memory caching (boolean),
    *     'onlyMemoryCaching' => enable / disable only memory caching (boolean),
    *     'memoryCachingLimit' => max nbr of records to store into memory caching (int),
    *     'fileNameProtection' => enable / disable automatic file name protection (boolean),
    *     'automaticSerialization' => enable / disable automatic serialization (boolean),
    *     'automaticCleaningFactor' => distable / tune automatic cleaning process (int),
    *     'hashedDirectoryLevel' => level of the hashed directory system (int),
    *     'hashedDirectoryUmask' => umask for hashed directory structure (int),
    *     'errorHandlingAPIBreak' => API break for better error handling ? (boolean)
    * );
    *
    * If sys_get_temp_dir() is available and the
    * 'cacheDir' option is not provided in the
    * constructor options array its output is used
    * to determine the suitable temporary directory.
    *
    * @see http://de.php.net/sys_get_temp_dir
    * @see http://pear.php.net/bugs/bug.php?id=18328
    *
    * @param array $options options
    * @access public
    */
    function __construct($options = array(NULL))
    {
        foreach($options as $key => $value) {
            $this->setOption($key, $value);
        }
        if (!isset($options['cacheDir']) && function_exists('sys_get_temp_dir')) {
          $this->setOption('cacheDir', sys_get_temp_dir() . DIRECTORY_SEPARATOR);
        }
    }

    /**
    * Generic way to set a Cache_Lite option
    *
    * see Cache_Lite constructor for available options
    *
    * @var string $name name of the option
    * @var mixed $value value of the option
    * @access public
    */
    function setOption($name, $value)
    {
        $availableOptions = array('errorHandlingAPIBreak', 'hashedDirectoryUmask', 'hashedDirectoryLevel', 'automaticCleaningFactor', 'automaticSerialization', 'fileNameProtection', 'memoryCaching', 'onlyMemoryCaching', 'memoryCachingLimit', 'cacheDir', 'caching', 'lifeTime', 'fileLocking', 'writeControl', 'readControl', 'readControlType', 'pearErrorMode');
        if (in_array($name, $availableOptions)) {
            $property = '_'.$name;
            $this->$property = $value;
        }
    }

    var $_WLock;
    var $_refreshTimeTolerance = 60;

    // write lock is always exclusive (only my thread and nobody else)
    function getWLock($block) {
      $lock = $block ? LOCK_EX : LOCK_EX|LOCK_NB;

      $lfile = $this->_file .".wlock";
      $fp = @fopen($lfile, "wb");
      if (!$fp) return null;

      // get exclusive lock
      $ok = @flock($fp, $lock);
      if ($ok) return $fp;

      @fclose($fp);
      return null;
    }
    // read lock is always blocking (always waiting for lock)
    function getRLock($exclusive) {
      $lock = $exclusive ? LOCK_EX : LOCK_SH;

      $lfile = $this->_file .".rlock";
      $fp = @fopen($lfile, "wb");
      if (!$fp) return null;

      // get shared (read) or ex. lock
      $ok = @flock($fp, $lock);
      if ($ok) return $fp;

      @fclose($fp);
      return null;
    }
    // unlock exclusive or shared lock
    function releaseLock($fp) {

      if (!$fp) return;
      if ($fp === "WLock") {
        $fp = $this->_WLock;
        $this->_WLock = null;
      }

      $ok = @flock($fp, LOCK_UN);
      $ok = @fclose($fp);
      return $ok;
    }

    /**
    * Test if a cache is available and (if yes) return it
    *
    * @param string $id cache id
    * @param string $group name of the cache group
    * @param boolean $doNotTestCacheValidity if set to true, the cache validity won't be tested
    * @return string data of the cache (else : false)
    * @access public
    */
    function &get($id, $group = 'default', $doNotTestCacheValidity = false)
    {
        // global $FIRE;
        $this->_id = $id;
        $this->_group = $group;
        $data = false;
        if ($this->_caching) {
            $this->_setRefreshTime();
            $this->_setFileName($id, $group);

            clearstatcache();
            $filemtime = file_exists($this->_file) ? @filemtime($this->_file) : 0;
            $this->_WLock = $this->getWLock($block=0);

            // write lock? (right to check cache validity and opportunity to write new cache content)
            if ($this->_WLock) {
              $isCacheValid = ($filemtime > $this->_refreshTime);
              // $FIRE->trace('$isCacheValid', $isCacheValid, convertDate($filemtime, "mktime", "H:i:s"), convertDate($this->_refreshTime, "mktime", "H:i:s"), date("H:i:s") );
              if ($isCacheValid) {
                $data = file_get_contents($this->_file);
                // $this->releaseLock($this->_WLock); $this->_WLock = null;
                $this->releaseLock("WLock");
              }
            }
            //
            else {
              $isCacheReadValid = (($filemtime + $this->_refreshTimeTolerance) > $this->_refreshTime);
              // wait for valid cache
              if (!$isCacheReadValid) {
                $this->releaseLock( $this->getWLock($block=1) );
              }

              $rlock = $this->getRLock($exclusive=0);
              $data = file_get_contents($this->_file);
              $this->releaseLock($rlock);
            }

//             if ($doNotTestCacheValidity || is_null($this->_refreshTime)) {
//                 if (file_exists($this->_file)) $data = $this->_read();
//             }
//             else {
//                 if (file_exists($this->_file) && (@filemtime($this->_file) > $this->_refreshTime)) $data = $this->_read();
//             }
            if ($this->_automaticSerialization) $data = unserialize($data);
            // $FIRE->trace('$this', $this);
            // $FIRE->trace('$data', $data);
            return $data;
        }
        return false;
    }

    /**
    * Save some data in a cache file
    *
    * @param string $data data to put in cache (can be another type than strings if automaticSerialization is on)
    * @param string $id cache id
    * @param string $group name of the cache group
    * @return boolean true if no problem (else : false or a PEAR_Error object)
    * @access public
    */
    function save($data, $id = NULL, $group = 'default')
    {
        // this method shouldn't be called wo obtained write lock
        if (!$this->_WLock) return false;

        if ($this->_caching) {
            if ($this->_automaticSerialization) $data = serialize($data);
            if (isset($id)) $this->_setFileName($id, $group);

            if ($this->_automaticCleaningFactor>0 && ($this->_automaticCleaningFactor==1 || mt_rand(1, $this->_automaticCleaningFactor)==1)) {
              $this->clean(false, 'old');
            }

            // get exclusive read lock
            $rlock = $this->getRLock($exclusive=1);
            // can write to file now
            $res = $this->_write($data);
            // release exclusive read lock
            $this->releaseLock($rlock);

            // release exclusive write lock obtained in $this->get()
            // $this->releaseLock($this->_WLock); $this->_WLock = null;
            $this->releaseLock("WLock");


            if (is_object($res)) {
                // $res is a PEAR_Error object
                if (!($this->_errorHandlingAPIBreak)) {
                    return false; // we return false (old API)
                }
            }
            return $res;

        }
        return false;
    }

    /**
    * Remove a cache file
    *
    * @param string $id cache id
    * @param string $group name of the cache group
    * @param boolean $checkbeforeunlink check if file exists before removing it
    * @return boolean true if no problem
    * @access public
    */
    function remove($id, $group = 'default', $checkbeforeunlink = false)
    {
        $this->_setFileName($id, $group);

        if ( $checkbeforeunlink ) {
            if (!file_exists($this->_file)) return true;
        }
        return $this->_unlink($this->_file);
    }

    /**
    * Clean the cache
    *
    * if no group is specified all cache files will be destroyed
    * else only cache files of the specified group will be destroyed
    *
    * @param string $group name of the cache group
    * @param string $mode flush cache mode : 'old', 'ingroup', 'notingroup',
    *                                        'callback_myFunction'
    * @return boolean true if no problem
    * @access public
    */
    function clean($group = false, $mode = 'ingroup')
    {
        return $this->_cleanDir($this->_cacheDir, $group, $mode);
    }

    /**
    * Set to debug mode
    *
    * When an error is found, the script will stop and the message will be displayed
    * (in debug mode only).
    *
    * @access public
    */
    function setToDebug()
    {
        $this->setOption('pearErrorMode', CACHE_LITE_ERROR_DIE);
    }

    /**
    * Set a new life time
    *
    * @param int $newLifeTime new life time (in seconds)
    * @access public
    */
    function setLifeTime($newLifeTime)
    {
        $this->_lifeTime = $newLifeTime;
        $this->_setRefreshTime();
    }

    /**
    * Return the cache last modification time
    *
    * BE CAREFUL : THIS METHOD IS FOR HACKING ONLY !
    *
    * @return int last modification time
    */
    function lastModified()
    {
        return @filemtime($this->_file);
    }

    /**
    * Trigger a PEAR error
    *
    * To improve performances, the PEAR.php file is included dynamically.
    * The file is so included only when an error is triggered. So, in most
    * cases, the file isn't included and perfs are much better.
    *
    * @param string $msg error message
    * @param int $code error code
    * @access public
    */
    function raiseError($msg, $code)
    {
        $this->_pearErrorMode = array($msg, $code);
        return false;
        // include_once('PEAR.php'); // facepalm
        // return PEAR::raiseError($msg, $code, $this->_pearErrorMode);
    }

    /**
     * Extend the life of a valid cache file
     *
     * see http://pear.php.net/bugs/bug.php?id=6681
     *
     * @access public
     */
    function extendLife()
    {
        @touch($this->_file);
    }

    // --- Private methods ---

    /**
    * Compute & set the refresh time
    *
    * @access private
    */
    function _setRefreshTime()
    {
        if (is_null($this->_lifeTime)) $this->_refreshTime = null;
        else                           $this->_refreshTime = time() - $this->_lifeTime;
    }

    /**
    * Remove a file
    *
    * @param string $file complete file path and name
    * @return boolean true if no problem
    * @access private
    */
    function _unlink($file)
    {
        if (!@unlink($file)) return $this->raiseError('Cache_Lite : Unable to remove cache !', -3);
        return true;
    }

    /**
    * Recursive function for cleaning cache file in the given directory
    *
    * @param string $dir directory complete path (with a trailing slash)
    * @param string $group name of the cache group
    * @param string $mode flush cache mode : 'old', 'ingroup', 'notingroup',
                                             'callback_myFunction'
    * @return boolean true if no problem
    * @access private
    */
    function _cleanDir($dir, $group = false, $mode = 'ingroup')
    {
        if ($this->_fileNameProtection) $motif = $group ? 'cache_'.md5($group).'_' : 'cache_';
        else $motif = $group ? 'cache_'.$group.'_' : 'cache_';


        if (!($dh = opendir($dir))) {
            return $this->raiseError('Cache_Lite : Unable to open cache directory !', -4);
        }

        $result = true;
        while (($file = readdir($dh)) !== false) {
            if (($file != '.') && ($file != '..')) {
                if (substr($file, 0, 6)=='cache_') {
                    $file2 = $dir . $file;
                    if (is_file($file2)) {
                        switch (substr($mode, 0, 9)) {
                            case 'old':
                                // files older than lifeTime get deleted from cache
                                if (!is_null($this->_lifeTime)) {
                                    if ((time() - @filemtime($file2)) > $this->_lifeTime) {
                                        $result = ($result and ($this->_unlink($file2)));
                                    }
                                }
                                break;
                            case 'notingrou':
                                if (strpos($file2, $motif) === false) {
                                    $result = ($result and ($this->_unlink($file2)));
                                }
                                break;
                            case 'callback_':
                                $func = substr($mode, 9, strlen($mode) - 9);
                                if ($func($file2, $group)) {
                                    $result = ($result and ($this->_unlink($file2)));
                                }
                                break;
                            case 'ingroup':
                            default:
                                if (strpos($file2, $motif) !== false) {
                                    $result = ($result and ($this->_unlink($file2)));
                                }
                                break;
                        }
                    }
                    if ((is_dir($file2)) and ($this->_hashedDirectoryLevel>0)) {
                        $result = ($result and ($this->_cleanDir($file2 . '/', $group, $mode)));
                    }
                }
            }
        }
        return $result;
    }

    /**
    * Make a file name (with path)
    *
    * @param string $id cache id
    * @param string $group name of the group
    * @access private
    */
    function _setFileName($id, $group)
    {

        if ($this->_fileNameProtection) $suffix = 'cache_'.md5($group).'_'.md5($id);
        else                            $suffix = 'cache_'.$group.'_'.$id;

        $root = $this->_cacheDir;
        if ($this->_hashedDirectoryLevel>0) {
            $hash = md5($suffix);
            for ($i=0 ; $i<$this->_hashedDirectoryLevel ; $i++) {
                $root = $root . 'cache_' . substr($hash, 0, $i + 1) . '/';
            }
        }
        $this->_fileName = $suffix;
        $this->_file = $root.$suffix;
    }

    /**
    * Write the given data in the cache file
    *
    * @param string $data data to put in cache
    * @return boolean true if ok (a PEAR_Error object else)
    * @access private
    */
    function _write($data)
    {
        if ($this->_hashedDirectoryLevel > 0) {
            $hash = md5($this->_fileName);
            $root = $this->_cacheDir;

            for ($i=0 ; $i<$this->_hashedDirectoryLevel ; $i++) {
                $root = $root . 'cache_' . substr($hash, 0, $i + 1) . '/';
                if (!@is_dir($root)) @mkdir($root, $this->_hashedDirectoryUmask);
            }
        }

        $fp = @fopen($this->_file, "wb");
        if ($fp) {
//             if ($this->_fileLocking) @flock($fp, LOCK_EX);
//             if ($this->_readControl) @fwrite($fp, $this->_hash($data, $this->_readControlType), 32);

//             $mqr = get_magic_quotes_runtime();
//             if ($mqr) set_magic_quotes_runtime(0);
            @fwrite($fp, $data);
//             if ($mqr) set_magic_quotes_runtime($mqr);

//             if ($this->_fileLocking) @flock($fp, LOCK_UN);

            // flush!
            @fflush($fp);
            @fclose($fp);
            return true;
        }
        return $this->raiseError('Cache_Lite : Unable to write cache file : '.$this->_file, -1);
    }

    /**
    * Make a control key with the string containing datas
    *
    * @param string $data data
    * @param string $controlType type of control 'md5', 'crc32' or 'strlen'
    * @return string control key
    * @access private
    */
    function _hash($data, $controlType)
    {
        switch ($controlType) {
        case 'md5':
            return md5($data);
        case 'crc32':
            return sprintf('% 32d', crc32($data));
        case 'strlen':
            return sprintf('% 32d', strlen($data));
        default:
            return $this->raiseError('Unknown controlType ! (available values are only \'md5\', \'crc32\', \'strlen\')', -5);
        }
    }

}
